#include "BasicSolver.hpp"
#include <algorithm>



BasicSolver::BasicSolver(Settings& sets, Statistics& stats):
	ConstraintDB                (sets, stats),
    var_act_hp                  ( sets.var_max_act , sets.var_min_inc, sets.var_decay),
    state_ok                    ( true  ),
    queue_begin                 ( 0     ),
    simplify_assign_idx         ( 0     ),
    simplify_props_remaining    ( 0     ),
    clause_stamp				( 0 	),
	currentId					( 0     )
{

	sapplifyalot = new Simplifier(this, &stats, settings.simpblocked, settings.simpvarelim, settings.printmodel);
}


BasicSolver::~BasicSolver(){}

/*
TrBool BasicSolver::isSAT(  uint32_t timeout){

	if (!state_ok)          return FALSE_TrBool;

    var_flags		.ensureSize(numberVariables());
    conflict_count	.setNumberClauses((double)numberAllClauses());
    luby_restarts	.setScale(128);

    if(settings.console_out != NULL ){
        settings.console_out->writeString(
                "|                           SApperloT (base)... YOU GOT ME SEARCHING ... ");
        settings.console_out->writeCenteredToBG(basename(settings.cnf_file_name));
        statistics.printStatsProgressHead();
        settings.console_out->writeString(
                "|____________________________________________________________________________________________|\n");
    }

   	TrBool status = UNKNOWN_TrBool;
   	while(statistics.cpuUserTime() <= timeout){

   		uint64_t conflict_max = luby_restarts.nextRestartGetConflictCount();

   		status = search( INIT, conflict_count.getConflicts() + conflict_max);

    	if(settings.console_out != NULL){ statistics.printStatsProgress(numberLiterals(),conflict_count.getConflicts());}

       	if(status.isKnown()) break;

       	backjumpKeepLevel(0,BACKJUMP_SET_PREV);
   	}


    if(status.isUnknown()){
    	asrt(timeout >= 0 && timeout <= statistics.cpuUserTime());
    	if(settings.console_out != NULL)
    		settings.console_out->writeString("|!!! Timeout of %i seconds exceeded !!! ",(int)timeout);
    	    return UNKNOWN_TrBool;
    }

    if(settings.console_out != NULL){
        settings.console_out->writeString("|____________________________________________________________________________________________");
        settings.console_out->writeString("| ");
    }



    asrt(!status.isTrue() || debugVerifyModel());

    //backjumpKeepLevel(0,BACKJUMP_SET_PREV);
    return status;
}
*/



void	BasicSolver::takePartialModelAsUnits(){
	asrt(state_ok);

	Variable x = INVALID_VAR;
	decision_level_end.clear();

	for(queue_begin	= 0;queue_begin < assign_stack.size();++queue_begin){
		varAssignLitAt0(assign_stack[queue_begin]);
	}

	if(++clause_stamp > Clause::MAX_STAMP){
		clsClearAllStamps();
		clause_stamp = 0;
		++statistics.nbr_expensive_cls_stamp;
	}

	simplify();


}


TrBool BasicSolver::extendModel(){
	if (state_ok)        return FALSE_TrBool;

   	TrBool status = search( INIT, max64);

   	asrt(status.isTrue()); asrt(debugVerifyModel());
    return status;
}



bool BasicSolver::addClause(vector<Literal>& lits){

    asrt(decisionLevel() == 0);

    if (!state_ok || lits.empty()){                                     return false;}

    // Check if clause is satisfied and remove false/duplicate literals:
    std::sort(lits.begin(),lits.end());

    Literal                     last    (INVALID_LIT);
    uint32_t                    i       (0);
    vector<Literal>::iterator   to      (lits.begin());
    while(i < lits.size()){
        if(litIsTrue(lits[i]) || lits[i] == last.flipped()){       		return true;}

        if(litIsUnknown(lits[i]) && lits[i] != last){
           *to++ = last = lits[i];
           maybeNewVariable(last.toVar());}

        ++i;
    }

    return uncheckedAddClause(lits.begin(),to);
}

bool BasicSolver::addClauseForSimplifier(vector<Literal>& lits){

    asrt(decisionLevel() == 0);

    if (!state_ok || lits.empty()){                                     return false;}

    return (*sapplifyalot).addClause( lits , ++currentId );
}

bool BasicSolver::addClauseFromSimplifier(Clause* myClause)
{
	vector<Literal> lits;
	lits.reserve((*myClause).size());

	for(uint i=0;i<(*myClause).size();i++)
		lits.push_back((*myClause)[i]);

	asrt(lits.size() == (*myClause).size());
	addClause(lits);

	return true;
}

bool BasicSolver::uncheckedAddClause  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend){


    uint32_t l_sz(lend-lbeg);

    if(l_sz == 0){                                               	return state_ok = false;}

    if(l_sz == 1){
        asrt(litIsUnknown(*lbeg));
        uncheckedEnqueue<Clause*>(*lbeg);                         	return (state_ok = (bcp(RETURN) == BCP_SUCCESS));}

    if(l_sz == 2){
        clsCreateBinary(*lbeg,*(lbeg+1));                   		return true;}

    asrt(l_sz <= Clause::MAX_SIZE);

    clsCreate(lbeg,lend);											return true;
}

void 	BasicSolver::reserveForClausesSimplifier(uint32_t need_sz)
{
	reserveForClauses(need_sz);
	if(settings.simp)
		(*sapplifyalot).reserveSpaceForClauses(need_sz);
}


Clause* BasicSolver::uncheckedAddClauseForSimplifier  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend){


    // int32_t id(clauses.size());
    uint32_t cl_sz							= lend-lbeg;

      //TODO:
  //    asrt(cl_sz > 2);

      void* mem								= malloc(sizeof(Clause) + sizeof(Literal)*cl_sz);
      if(mem == NULL){
    	fprintf(stderr,"c !!!OUT OF MEMORY - Try to Fix!!!\n");
      	exit(EXIT_FAILURE);
      	return NULL;
      }
      Clause* cls 							= new (mem) Clause(cl_sz, false, ++currentId);

//      Literal lit0 							= *lbeg;
//      Literal lit1 							= *(lbeg+1);
      for(uint32_t i = 0;lbeg != lend;++lbeg) {(*cls)[i++] = *lbeg;}


      return cls;
}

TrBool BasicSolver::search(const OnConflict on_conflict,uint64_t abs_max_confl ){

    asrt(state_ok);
    TrBool              result				= UNKNOWN_TrBool;
    uint32_t 			learnts_allowed;

    ++statistics.nbr_restarts;

    while(conflict_count.getConflicts() < abs_max_confl || decisionLevel() == 0 ){

        int32_t bcp_res    =     bcp(on_conflict);

        if(bcp_res == BCP_SUCCESS){         //there is no conflict

             // Simplify the set of problem clauses
            if (decisionLevel() == 0    && (result = simplify(), result.isKnown())) {               return result;        }

            // Reduce the set of learnt clauses:
			learnts_allowed = conflict_count.getMaxLearnts() + this->numberAssignments();
			if(learnts_allowed <= numberLearntLongClses()){learntsGC((learnts_allowed * 3)/4); }


            // New decision:
            Literal next    = nextDecision();

            //check if model is found
            if (next == INVALID_LIT)    {                                                           return TRUE_TrBool;   }

            // Increase decision level and enqueue 'next'
            asrt(queue_begin == assign_stack.size());
            asrt(litIsUnknown(next));
            newDecisionLevel();
            uncheckedEnqueue<Clause*>(next);

            //if(decisionLevel() == 1){firstDecisionCheckBothPol(next);}

            set_max_int(statistics.max_dec_level,decisionLevel());


        }else{                         //there is a conflict

        	conflict_count.incConflict();

			asrt(bcp_res != 1);
			if (!firstUIP(bcp_res))  {                                                             	return TrBool(state_ok = false); }
			asrt(litAssignLevel(conflict_lemma[0]) == decisionLevel());

			reduceSelfImpl();

			uint32_t backtrack_level = backtrackLevel();

			int32_t bj_dist = decisionLevel() - backtrack_level;
			statistics.storeBackjumpingDistance(bj_dist);

			backjumpKeepLevel(backtrack_level,BACKJUMP_SET_PREV);
			asrt(litIsUnknown(conflict_lemma[0]));

			const uint32_t l_sz(conflict_lemma.size());
			if(l_sz > 2){
			   asrt(l_sz <= Clause::MAX_SIZE);
			   Clause* learnt_cl = clsCreate(conflict_lemma.begin(),conflict_lemma.end(),true);
			   clsBump(learnt_cl);
			   uncheckedEnqueue(conflict_lemma[0], learnt_cl);
			}else if(l_sz == 2){
			   clsCreateBinary(conflict_lemma[0],conflict_lemma[1]);
			   uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1]);
			}else{
			   uncheckedEnqueue<Clause*>(conflict_lemma[0]);
			}


			//panic too many literals

			if ( numberLiterals() > settings.max_literals){
				//print_out("Force Learnts GC due to Memory: %llu Literals, and %llu allowed\n",numberLiterals(),settings.max_literals);
				learntsGC(numberLearntLongClses() / 3);
				if(on_conflict != UNLINK){
					backjumpKeepLevel(0,BACKJUMP_SET_PREV_FLP);
					break;}
			}

			var_act_hp.decayAll();
			claDecayActivity();

        }
    }

    asrt(on_conflict != UNLINK);
    //maximal number of conflicts reached
    return UNKNOWN_TrBool;

}

void    BasicSolver::backjumpUndoAssigns (uint32_t assign_idx,const BackjumpType type){

	Variable x = INVALID_VAR;
	asrt(assign_stack.size() >= assign_idx);

	queue_begin 	= assign_stack.size();

	if(type == BACKJUMP_SET_PREV){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSetPrev(x);
			reinsertVarOrder(x);}
	}else if(type == BACKJUMP_SET_PREV_FLP){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSetPrevFlp(x);
			reinsertVarOrder(x);}
	/*}else if(type == BACKJUMP_SET_PREV_RAND){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			settings.randy.nextBool() ? varClearAssignmentSetPrevFlp(x):varClearAssignmentSetPrev(x) ;
			reinsertVarOrder(x);}*/
	}else{
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSimple(x);
			reinsertVarOrder(x);}
	}

	assign_stack.resize(queue_begin);


	if(++clause_stamp > Clause::MAX_STAMP){
		clsClearAllStamps();
		clause_stamp = 0;
		++statistics.nbr_expensive_cls_stamp;
	}
}


Literal BasicSolver::nextDecision(){

    Variable var = INVALID_VAR;

    if(settings.randy.nextDouble() < settings.random_var_freq){
        //random decision
        do{
            if(var_act_hp.empty()) return INVALID_LIT;
            var     = var_act_hp.removeAtPos(settings.randy.nextInt(var_act_hp.size()));

        }while(varValue(var).isKnown());

        ++statistics.nbr_rnd_decisions;

    }else{
        //activity based decision
        do{
            if(var_act_hp.empty()) return INVALID_LIT;
            var     = var_act_hp.removeMax();
        }while(varValue(var).isKnown() );

        ++statistics.nbr_act_decisions;
    }


    asrt(var != INVALID_VAR );

    switch (settings.polarity_mode){
        case Polarity_False     : return var.toNegLit();
        case Polarity_True      : return var.toPosLit();
        case Polarity_Rand      : return settings.randy.nextBool() ?
                                     var.toNegLit():var.toPosLit();
        case Polarity_Prev      : return varPrevValue(var).isTrue() ?
                                    var.toPosLit() : var.toNegLit();
        case Polarity_Hook      : return varHook(var).isTrue() ?
                                           var.toPosLit() : var.toNegLit();
        case Polarity_FlpHook   : return varHook(var).isFalse() ?
                                                   var.toPosLit() : var.toNegLit();
        default: asrt(false); return var.toNegLit();
    }
}

void	BasicSolver::firstDecisionCheckBothPol(Literal next){
	const uint32_t impl_start = assign_stack.size();
	asrt(assign_stack[impl_start-1] == next);
	asrt(tmp_literal_store.empty());

	if(BCP_SUCCESS == bcp(RETURN)){
		if(impl_start == assign_stack.size()){return;}

		for(uint32_t i = assign_stack.size();i != impl_start;){
			tmp_literal_store.push_back(assign_stack[--i]);};

		backjumpUndoAssigns(impl_start-1,BACKJUMP_PREV_UNTOUCHED);
		uncheckedEnqueue<Clause*>(next.flipped());
		asrt(decisionLevel() == 1);
		if(BCP_SUCCESS == bcp(RETURN)){
			uint32_t to = 0;

			for(uint32_t i = 0; i != tmp_literal_store.size();++i){
				Literal lit = tmp_literal_store[i];
				if(litIsTrue(lit)){tmp_literal_store[to++] = lit;}
				else if(litIsFalse(lit)){
					if(!containsBinary(next.flipped(),lit)){
						clsCreateBinary(next.flipped(),lit);
						//learnt_bin_clauses.push_back(std::pair<Literal,Literal>(next.flipped(),lit));
					}
					if(!containsBinary(next,lit.flipped())){
						clsCreateBinary(next,lit.flipped());
						//learnt_bin_clauses.push_back(std::pair<Literal,Literal>(next,lit.flipped()));
					}
				}
			}

			if(to != 0){
				backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);
				statistics.nbr_flipped_prop_learnts += to;
				while(to--){uncheckedEnqueue<Clause*>(tmp_literal_store[to]);}
			}else{
				backjumpUndoAssigns(impl_start-1,BACKJUMP_PREV_UNTOUCHED);
				uncheckedEnqueue<Clause*>(next);
			}

		}else{
			backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);
			++statistics.nbr_flipped_prop_learnts;
			uncheckedEnqueue<Clause*>(next);
		}
	}else{
		backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);
		++statistics.nbr_flipped_prop_learnts;
		uncheckedEnqueue<Clause*>(next.flipped());
	}
	tmp_literal_store.clear();
}




int32_t	BasicSolver::bcp(const OnConflict on_conflict){

    int32_t        	nbr_pivot_lits = BCP_SUCCESS; //indicates no conflict by now
    uint32_t        num_props(0);

    while(queue_begin < assign_stack.size()){

        ++num_props;
        nbr_pivot_lits = bcpPropagateLit(on_conflict,assign_stack[queue_begin++].flipped());

    }

    statistics.nbr_propagations    += num_props;
    simplify_props_remaining        -= num_props;

    return nbr_pivot_lits;
}


int32_t	BasicSolver::bcpPropagateLit(const OnConflict on_conflict, const Literal& false_lit){

	 we_vec& watched    = litWatchedClauses(false_lit); //watched clauses that have to be checked when lit gets true
	 we_iter iter = watched.begin();
	 we_iter copy = iter;
     we_iter end  = watched.end();

     while(iter != end){
    	 WatchedElement& we = *iter++;

		//might be any other literal of this clause except the watched literal
        const Literal any_other = we.getLiteral();
        asrt(any_other != false_lit);

    	 //keep this watcher if other lit is true
    	 if(litIsTrue(any_other)){ *copy++ = we; continue;}

    	 if(we.isBinary()){
			*copy++ = we; //keep this watcher
			if(litIsUnknown(any_other)){
				uncheckedEnqueue<const Literal&>(any_other,false_lit);
				continue;
			}else{ //other literal in binary clause is false --> conflict
				bcpFinishOnConflict(watched,iter,copy,end);
				int nbr_pivot_lits = (on_conflict == RETURN) ? 1 : bcpInitConflict(decisionLevel(),false_lit,any_other);
				asrt(nbr_pivot_lits != 1 || (on_conflict != INIT));
				return nbr_pivot_lits;
			}
    	 }


		Clause& cl = *(we.getClause());

        uint32_t other_watcher_index = (cl[1] != false_lit);
        const Literal other_w = cl[other_watcher_index];
        asrt(cl[other_watcher_index ^ 1] == false_lit && other_w != false_lit);

        //it might be that the shortcut literal in 'we' does not point to the actual watched literal anymore
        if(litIsTrue(other_w)){
            we.setLiteral(other_w);
            *copy++ = we; //keep this watcher
            continue;
        }

		if(bcpFindNewWatcher(cl,false_lit,other_watcher_index)){continue;}


		// No new watched found -- clause is unit under assignment:
	    we.setLiteral(other_w);

	    if (litIsFalse(other_w)){

	    	clsBump(&cl);

	    	if(on_conflict == UNLINK){
	    		//unlink this clause and continue
	    		clsIrregularUnlink(&cl,litWatchedClauses(other_w));
	    		continue;
	    	}

	    	*copy++	= we;
	    	bcpFinishOnConflict(watched,iter,copy,end);

	    	if(on_conflict == INIT){
	    		const int32_t dec_level = decisionLevel();
	    		int32_t  nbr_pivot_lits = bcpInitConflict(dec_level,false_lit,other_w);
	            for(uint32_t k = 2;k<cl.size();++k){firstUIPCheckLiteral(cl[k],dec_level,nbr_pivot_lits);}
	            asrt(nbr_pivot_lits != 1);
	            return nbr_pivot_lits;
	        }else{
	        	asrt(on_conflict == RETURN);
	        	return 1; //any positive value
	        }

	    }else{
	    	*copy++ = we;
	    	uncheckedEnqueue(other_w, &cl);
	    }

	 }
     if(copy != end) watched.erase(copy,end);

     return BCP_SUCCESS;
}




bool BasicSolver::firstUIP(int32_t nbr_pivot_elements){
	asrt(nbr_pivot_elements >= 0);
	if(nbr_pivot_elements == 0) return false;

    const uint32_t   dec_level           (decisionLevel());
    uint32_t        assign_stack_index  = assign_stack.size() ;

    asrt(nbr_pivot_elements != 1);
    AssignInfo 	confl_info;
    Literal 	pivot_lit;

    while(true){

    	//skip all unseen assignments
    	//since there always exists a UIP the index cannot become negative
    	while (!var_flags.isSet(assign_stack[--assign_stack_index].toVar().idx())){}
    	asrt(assign_stack_index >= 0);

    	pivot_lit                       = assign_stack[assign_stack_index];
    	//can be already cleared since this lit cannot appear in any remaining (previous) reason because literals
    	//in reason clauses are all false (except for the one it is the assign reason)
    	var_flags                       .clear(pivot_lit.toVar().idx());

    	if(--nbr_pivot_elements == 0){
    		asrt(litAssignLevel(pivot_lit) == dec_level);
    		asrt(pivot_lit != INVALID_LIT);
    		conflict_lemma[0]                = pivot_lit.flipped();
    		return true;
    	}

    	confl_info                      = varAssignInfo(pivot_lit.toVar());
    	asrt(litAssignLevel(pivot_lit) == dec_level);

    	asrt(!confl_info.noReason());

    	if(confl_info.isBinReason()){
    		firstUIPCheckLiteral(confl_info.getBinReason(),dec_level,nbr_pivot_elements);
        	//settings.console_out->writeString("Short Reason: %d %d ",pivot_lit.toDIMACS(),confl_info.getShortFirstReason().toDIMACS());
    	}else{
    		asrt(confl_info.isClsReason() && confl_info.getClsReason() != NULL);

    		Clause& cl                      = *confl_info.getClsReason();
    	    const uint32_t cl_sz            (cl.size());
    	    /*settings.console_out->writeString("Reason Clause: ");
    	    cl.print(*settings.console_out);*/

    	    asrt(pivot_lit == cl[0]|| pivot_lit == cl[1]);
    	    asrt(litAssignLevel(cl[0]) == dec_level && litAssignLevel(cl[1]) == dec_level);

    	    clsBump(&cl);

    	    uint32_t cl_index = pivot_lit == cl[0];//pivot_lit ==cl[0] ? 1 : 0
    	    asrt(cl[cl_index ^1] == pivot_lit);

    	    //do for all except the pivot_lit
    	    firstUIPCheckLiteral(cl[cl_index],dec_level,nbr_pivot_elements);
    	    for(cl_index = 2;cl_index < cl_sz;++cl_index){
    	    	firstUIPCheckLiteral(cl[cl_index],dec_level,nbr_pivot_elements); }
    	}
	}
    return false;
}

uint32_t BasicSolver::backtrackLevel(){
    asrt(litAssignLevel(conflict_lemma[0]) == decisionLevel());
    uint32_t lsz = conflict_lemma.size();
    if(lsz < 2) return 0;

    uint32_t max_idx = 1;
    while(--lsz > 1){
        if(litAssignLevel(conflict_lemma[lsz]) > litAssignLevel(conflict_lemma[max_idx])){max_idx = lsz;}}

    //such that the literals with maximum level are the watchers
    //this is important for backtracking!
    Literal highest				= conflict_lemma[max_idx];
    conflict_lemma[max_idx]     = conflict_lemma[1];
    conflict_lemma[1]           = highest;


    return litAssignLevel(conflict_lemma[1]);
}




void     BasicSolver::reduceSelfImpl(){

    //asrt(c_db.checkAssignReasons());

    if(conflict_lemma.size() < 2){asrt(var_flags.allCleared()); return;}

    sort(conflict_lemma.begin()+1,conflict_lemma.end(),LitActDescendingFunctor(this));
    asrt(var_flags_to_clear.empty());
    //var_flags.clearAll();

    uint cp,idx;
    for(cp = idx = 1; idx != conflict_lemma.size(); ++idx){
        if(litIsNecessary(conflict_lemma[idx])) {conflict_lemma[cp++] = conflict_lemma[idx];}}

    conflict_lemma.resize(cp);

    unsetConfLemmaFlags();
}


bool      BasicSolver::litIsNecessary(Literal lit){

    if(varAssignInfo(lit.toVar()).noReason()) return true;

    const Variable start(lit.toVar());

    uint32_t unmark_top(var_flags_to_clear.size());
    asrt(tmp_literal_store.empty());

    while(true){

        asrt(!varAssignInfo(lit.toVar()).noReason());
        const AssignInfo& a_info = varAssignInfo(lit.toVar());

        bool is_nec;

        if(a_info.isBinReason()){
            is_nec = reduceSelfImpCheckLiteral(a_info.getBinReason());

        }else{
            asrt(a_info.isClsReason() && a_info.getClsReason() != NULL);

            Clause& cl	= *(a_info.getClsReason());

            uint32_t cl_index = lit.toVar() == cl[0].toVar();
            asrt(cl[cl_index^1].toVar() == lit.toVar());
            asrt(var_flags.isSet(cl[cl_index^1].toVar().idx()));

            is_nec = reduceSelfImpCheckLiteral(cl[cl_index]);
            for(cl_index = 2;!is_nec && cl_index < cl.size();++cl_index){
                is_nec = reduceSelfImpCheckLiteral(cl[cl_index]);
            }
        }

        if(is_nec){
            for( ;unmark_top < var_flags_to_clear.size();++unmark_top)   {var_flags.clear(var_flags_to_clear[unmark_top].idx());}
            for(int i = tmp_literal_store.size();i--;)                   {var_flags.clear(tmp_literal_store[i].toVar().idx());}
            tmp_literal_store.clear();
            return true;
        }


        if(tmp_literal_store.empty()){
            statistics.total_lemma_shortening += 1;
            //shorten_lemma_to_unmark.clear();
            var_flags_to_clear.push_back(start);
            return false;
        }

        lit = tmp_literal_store.back(); tmp_literal_store.pop_back();
        var_flags_to_clear.push_back(lit.toVar());
    }
}

void BasicSolver::clearBinWatchers(Literal lit){

	asrt(litIsUnknown(lit));
	we_iter from,to,end;
	we_vec& wes = litWatchedClauses(lit);

	from = to = wes.begin(); end = wes.end();
	while(from != end){
		WatchedElement& we = *from++;
		if(we.isBinary() && litIsTrue(we.getLiteral())) {continue;} //delete watcher

		*to++ = we;
	}
	if(to != end){wes.erase(to,end);}
}



void BasicSolver::clearBinWatchers(Variable var){
	if(var_flags.isSet(var.idx()) || varValue(var).isKnown()) return;
	var_flags.setFlag(var.idx());	var_flags_to_clear.push_back(var);

	clearBinWatchers(var.toNegLit());
	clearBinWatchers(var.toPosLit());
}


void	BasicSolver::simplifyBinWatchersOf (we_vec& wes){
	we_iter from,to;

	for(from = to = wes.begin(); from != wes.end();++from){
		if(from->isBinary()){ clearBinWatchers(from->getLiteral().toVar());	}
		else{*to++ = *from;}		//might be since bigsimplify changes are not simplified further
	}
	wes.erase(to,wes.end());
}



void	BasicSolver::simplifyAllBinWatchers (){
	asrt(var_flags.allCleared());
	asrt(decisionLevel() == 0);

	while(simplify_assign_idx < assign_stack.size()){
		Literal lit = assign_stack[simplify_assign_idx++];
		simplifyBinWatchersOf(litWatchedClauses(lit));
		simplifyBinWatchersOf(litWatchedClauses(lit.flipped()));
	}


	for(uint32_t i = var_flags_to_clear.size();i--;){var_flags.carelessClear(var_flags_to_clear[i].idx());}
	var_flags_to_clear.clear();
	asrt(var_flags.allCleared());
}



TrBool BasicSolver::simplify( ){

    asrt(decisionLevel() == 0);

    if (!state_ok || bcp(RETURN) != BCP_SUCCESS){   return TrBool(state_ok = false);}

    if (assign_stack.size() != simplify_assign_idx  && (simplify_props_remaining <= 0)){

        ++statistics.nbr_simplify_execs;

        removeSat(learnt_cls);
        removeSat(fix_cls);

        simplifyAllBinWatchers();

        simplify_props_remaining   = numberLiterals();
    }

    return UNKNOWN_TrBool;
}



const char*  basic_solver_cpp_version(){return "BasicSolver.cpp:    $Revision: 1.4 $";}
