#include "Settings.hpp"
#include "BasicSolver.hpp"
#include "DMRPSolver.hpp"
#include "ConstraintDB.hpp"
#include "Types.hpp"
#include "Settings.hpp"
#include "IFStream.hpp"
#include "Settings.hpp"
#include "Random.hpp"
#include "BitVector.hpp"
#include "DecayHeap.hpp"



Settings::Settings():
    max_literals			  (120000000),
	var_decay              	  ( 135 , 1 ), //= 135/128
	var_min_inc				  (1, var_act_t::minNonZeroNumber().getLogOfDenominator() - (2* var_decay.getLogOfDenominator() )),
	var_max_act				  (1000),
    cls_decay              	  ( 129 , 1 ), //= 129/128
    cls_min_inc				  (1, cls_act_t::minNonZeroNumber().getLogOfDenominator() - (2* cls_decay.getLogOfDenominator() )),
    cls_max_act				  (1000),
    random_var_freq           ( 0.02 ),
    conflict_max_init         ( 100 ),
    conflict_max_inc          ( 1.5 ),
    learntsize_factor         ( (double)1.0 / (double)3.0),
    learntsize_inc            ( 1.05 ),
    cnf_file_name             ( NULL ),
    solver_name               ( NULL ),
    console_out				  ( NULL ),
    simpblocked				(false)
{
    Option<bool>::val_type val_true     = std::make_pair<const char*,bool>("yes",true);
    Option<bool>::val_type val_false    = std::make_pair<const char*,bool>("no",false);
    Option<bool>
    b_opt = Option<bool>(&print_console,    false,  "-console=",  	"If yes, some information is print to console output");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);

    b_opt = Option<bool>(&print_local,      true,  "-print-local=",    "If yes, local settings are used for printing (decimal comma)");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);

    b_opt = Option<bool>(&hybrid,    		true,  	"-hybrid=",  		"If yes, DMRP solving is performed for each short restart");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);

    b_opt = Option<bool>(&global_make_count,false,  "-global-mc=",  	"If yes, make count considers all unsat clauses not only temporarily unsat ones");
	b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);

    b_opt = Option<bool>(&printmodel,    	true,  "-printmodel=",  	"If yes and formula is SAT, satisfying variable assignments were printed");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);

    /*b_opt = Option<bool>(&remove_satisfied, true,  "-remove-satisfied=","If yes, satisfied clauses are removed");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    bool_options.push_back(b_opt);*/

    Option<char*>
    s_opt = Option<char*>(&csv_file_name,   NULL,  "-csv-out=",         "If specified, statistics are appended to this file");
    string_options.push_back(s_opt);


    Option<uint>
    i_opt = Option<uint>(&rand_seed,      91648253,  "-seed=",           "Specify random seed");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&timeout,           ~0,     "-timeout=",        "If positive, the solver terminates after the given number of seconds");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&simptimeout,           100,     "-simptimeout=",        "If positive, the simplifier terminates after the given number of seconds");
    int_options.push_back(i_opt);

    i_opt = Option<uint>(&polarity_mode, Polarity_Prev, "-polarity-mode=","To determine the polarity that is used to assign a decision variable");
    i_opt.addPossibleValue(std::make_pair<const char*,int>("false", Polarity_False  ));
    i_opt.addPossibleValue(std::make_pair<const char*,int>("true",  Polarity_True   ));
    i_opt.addPossibleValue(std::make_pair<const char*,int>("rand",  Polarity_Rand   ));
    i_opt.addPossibleValue(std::make_pair<const char*,int>("prev",  Polarity_Prev   ));
    int_options.push_back(i_opt);

    //simplifier:
    b_opt = Option<bool>(&simp, true, "-simp=", 			"If yes, simplification of the CNF formula is activated");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    simp_options.push_back(b_opt);

    /*b_opt = Option<bool>(&simpblocked, false, "-simpblocked=", "If yes, simplification searches for blocking clauses and temporarily deletes them from the formula to support subsumption.");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    simp_options.push_back(b_opt);*/

    b_opt = Option<bool>(&simpvarelim, true, "-simpvarelim=", "If yes, simplification tries to eliminate variables");
    b_opt.addPossibleValue(val_true);
    b_opt.addPossibleValue(val_false);
    simp_options.push_back(b_opt);

}

Settings::~Settings(){
	if(console_out != NULL){asrt(print_console);delete console_out;}
}


bool Settings::read(int argc, char** argv){

    //check for correct usage
    solver_name = bname(argv[0]);
    if(argc < 2){
    	fprintf(stderr,"c Not enough Parameters! \n");
        return false;
    }

    int index(0);
    char* value;
    command_line    = std::string(argv[index]);
    while(++index < argc){

        command_line    += std::string(argv[index]);

        for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){
            if((value = bool_options[opt_idx].canHandle(argv[index]))){
                if(bool_options[opt_idx].findValue(value))
                    goto OptionHandled; else return false;}
        }

        for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){
            if((value = int_options[opt_idx].canHandle(argv[index]))){
                if(!int_options[opt_idx].findValue(value)){
                    int_options[opt_idx].assign(atoi(value));}
                goto OptionHandled;
            }
        }

        for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){
            if((value = string_options[opt_idx].canHandle(argv[index]))){
                string_options[opt_idx].assign(value);
                goto OptionHandled;}
        }

        for(uint opt_idx = 0; opt_idx != simp_options.size();++opt_idx){
			if((value = simp_options[opt_idx].canHandle(argv[index]))){
				if(simp_options[opt_idx].findValue(value))
					 goto OptionHandled;
				else
					return false;
			}



		}

        if(*argv[index] == '-'){
        	fprintf(stderr,"c Unknown option: %s \n",argv[index]);
            return false;
        }else if(cnf_file_name != NULL){
        	fprintf(stderr,"c One of the two arguments is incorrect: \"%s\" or \"%s\" \n",cnf_file_name,argv[index]);
            return false;
        }else {
            cnf_file_name     = argv[index];
        }

        OptionHandled: ;
    }

    randy.setSeed(rand_seed);
    if(print_console){console_out = new OFStream(stderr);}

    return cnf_file_name != NULL;
};


void Settings::printUsage(){

	fprintf(stderr,"Program Usage: \n");
	fprintf(stderr,"%s [Options] <dimacs-file> \n",solver_name);
	fprintf(stderr,"Options are: \n");
    for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){    bool_options[opt_idx].printDescription();}
    for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printDescription();}
    for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){     int_options[opt_idx].printDescription();}
    for(uint opt_idx = 0; opt_idx != simp_options.size();++opt_idx){    simp_options[opt_idx].printDescription();}

    fprintf(stderr,"\n");

    fprintf(stderr,"Current Values are: \n");
    for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){    bool_options[opt_idx].printValue("%i");}
    for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printValue("%s");}
    for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){     int_options[opt_idx].printValue("%i");}
    for(uint opt_idx = 0; opt_idx != simp_options.size();++opt_idx){    simp_options[opt_idx].printValue("%i");}

   // printCVSVersionNumbers();
}




void Settings::printCVSVersionNumbers(){

	fprintf(stderr,"c LubyDMRP SApperloT\n");
#ifdef NDEBUG
    fprintf(stderr,"c CVS Version Numbers (RELEASE Compilation)\n");
#else
    fprintf(stderr,"c CVS Version Numbers (DEBUG Compilation)\n");
#endif

    fprintf(stderr,"c %s\n",basic_solver_cpp_version());
    fprintf(stderr,"c %s\n",constraintdb_cpp_version());
    fprintf(stderr,"c %s\n",dmrp_solver_cpp_version());
    fprintf(stderr,"c %s\n",sapperlot_cpp_version());
    fprintf(stderr,"c %s\n",settings_cpp_version());
    fprintf(stderr,"c %s\n",statistics_cpp_version());

    fprintf(stderr,"c %s\n",basic_solver_hpp_version);
    fprintf(stderr,"c %s\n",conflicts_restarts_hpp_version);
    fprintf(stderr,"c %s\n",constraintdb_hpp_version);
    fprintf(stderr,"c %s\n",dmrp_solver_hpp_version);
    fprintf(stderr,"c %s\n",settings_hpp_version);
    fprintf(stderr,"c %s\n",statistics_hpp_version);
    fprintf(stderr,"c %s\n",types_hpp_version);

    fprintf(stderr,"c %s\n",bitvector_hpp_version);
    fprintf(stderr,"c %s\n",decay_heap_hpp_version);
    fprintf(stderr,"c %s\n",ifstream_hpp_version);
    fprintf(stderr,"c %s\n",lazy_frac_hpp_version);
    fprintf(stderr,"c %s\n",ofstream_hpp_version);
    fprintf(stderr,"c %s\n",random_hpp_version);
    //fprintf(stderr,"c %s\n",info_hpp_version);

    //print_out("%s\n",utility_hpp_version);
}



const char*  settings_cpp_version(){return "Settings.cpp:       $Revision: 1.3 $";}
